unit dbase;

{......................................................................
  Unit for reading and writing dbase/foxbase/clipper databases.
  Copyright (C) 1998 Alejandro Vargas & Mauricio Ponzo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


Description:
-----------

  This unit is intended to give an easy way to use dbf compatible files with
  Pascal programs in order to make easy the administration of data files
  used by the program but created and manteined with foxbase i.e. We tryed to
  create methods that remembers the original dbase commands.
  

  ChangeLog
  ---------

  Current version is 0.1
   
  Version     Date             Remarks
  -------     ----             ----
  0.1         11/18/98         Basic managing of dbf files.
  0.2         02/12/01         Bugfix on appendblank function.
  0.3         09/24/01         Polimorphic function calls for fpc.
                               Changed RegistroActual by recno.
  0.4         07/31/03         Bugfix in value returned by locate function.
			       WARNING: now it returns the ioresult times (-1)
			                instead of ioresult.			       
                               Implement RecCount function.
  
To do:
-----
  Memo fields support.
  Index support.


Contact Us:
----------

  Alejandro Vargas
  anv@lanet.com.ar

  Mauricio Ponzo
  mponzo@xnetcuyo.com.ar

.......................................................................}





interface

const DBEOF=-1;
      DBBOF=0;

type Encabezado=record
     desc0: byte;
     anio: byte;
     mes: byte;
     dia: byte;
     CantReg: longint;
     LongEnc: word;
     LongReg: word;
     desc2: array[1..20] of char;
     end;


type FieldInfo=record
     nam:string[12];
     typ:char;
     Len:byte;
     Dec:byte;
     Off:word;
     end;


type dbf=object
     archivo:string;
     enc:encabezado;
     recno:longint;
     deleted:boolean;
     found:boolean;

     NFields:byte;
     Structure: array[1..255] of ^FieldInfo;

     function use(archi:string):integer;
     function eof:boolean;
     function bof:boolean;
     function go(registro:longint):longint;
     procedure gotop;
     procedure gobottom;
     function skip(registros:longint):longint;
     {$ifdef fpc}         {fpc supports polimorfic}
     function skip:longint;
     function leer(campo:string; var dato_string:string):integer;
     function leer(campo:string; var dato_real:real):integer;
     function leer(campo:string; var dato_longint:longint):integer;
     function leer(campo:string; var dato_bool:boolean):integer;
     function leer(campo:string):string;
     
     function grabar(campo:string;dato_string:string):integer;
     function grabar(campo:string;dato_real:real):integer;
     function grabar(campo:string;dato_longint:longint):integer;
     function grabar(campo:string;dato_bool:boolean):integer;
     
     function locate(campo:string;dato_string:string):integer;
     function locate(campo:string;dato_real:real):integer;
     function locate(campo:string;dato_longint:longint):integer;
     function locate(campo:string;dato_bool:boolean):integer;
     
     function locate(campo:string;dato_string:string;valorskip:longint):integer;
     function locate(campo:string;dato_real:real;valorskip:longint):integer;
     function locate(campo:string;dato_longint:longint;valorskip:longint):integer;
     function locate(campo:string;dato_bool:boolean;valorskip:longint):integer;

     {$else}
     function leer(campo:string;var dato):integer;
     function grabar(campo:string;var dato):integer;
     function locate(campo:string;var dato;valorskip:longint):integer;
     {$endif}
     procedure closeDatabase;
     function delete_mark(marca:char):integer;
     function delete:integer;
     function recall:integer;
     function appendblank:integer;
     function reccount:longint;

     end;


implementation

type DescripcionCampos=record
     Nombre: array[1..11] of char;
     Tipo: char;
     Desc3: array[1..4] of char;
     Longitud: byte;
     Decimales: byte;
     Desc4: array[1..14] of char;
     end;

var ultcampo:fieldinfo;

{-----------------------------------------------------------------------}
function c2pas(cadena:pchar;maximo:byte):string;
var a:byte;
    r:string;
begin
a:=0;
r:='';
while (cadena[a]<>#0) and (a<maximo)
do begin
   r:=r+cadena[a];
   inc(a);
   end;
c2pas:=r;
end;


{-----------------------------------------------------------------------}
function rtrim(entrada:string):string;
var a:byte;
    salida:string;
begin
a:=length(entrada);
while (a>0) and (entrada[a]=' ') do dec(a);
salida:=copy(entrada,1,a);
rtrim:=salida;
end;

{-----------------------------------------------------------------------}
function DBF.use(archi:string):integer;
{ Opens a database file and fills the info in the database object.      
  Returns the ioresult of the operation (0 if ok)                       }

var f:file;
    i:integer;
    offset:word;

    dsc:DescripcionCampos;
begin
     nFields:=0;
     {$I-}
     self.archivo:=archi;
     assign(f,self.archivo);
     reset(f,1);
     i:=ioresult;
     if i>0
     then begin
          use:=i;
          exit;
          end;

     blockread(f,self.enc,sizeof(self.enc));

     i:=ioresult;
     if i=0
     then begin
          if self.enc.CantReg=0
          then self.recno:=-1
          else self.recno:=1;
          self.deleted:=false;
          end
     else self.recno:=-1;
     use:=i;


     {Leer la estructura}

     NFields:=1;
     offset:=1;
     dsc.nombre[1]:=#0;
     while (nFields<=(self.enc.LongEnc-33) div 32) and (dsc.nombre[1]<>#$0d)
     do begin
        blockread(f,dsc,sizeof(dsc));
        i:=ioresult;
        if i>0
        then begin
             use:=-1;
             close(f);
             exit;
             end;

        new(Structure[nFields]);

        Structure[nFields]^.Nam:=rtrim(c2pas(@dsc.nombre,11));
        Structure[nFields]^.Typ:=dsc.tipo;
        Structure[nFields]^.Len:=dsc.longitud;
        Structure[nFields]^.Dec:=dsc.decimales;
        Structure[nFields]^.Off:=offset;

        inc(offset,dsc.longitud);

        inc(nFields);
        end;

dec(nFields);
close(f);
{$i+}
use:=0;
end;

{-----------------------------------------------------------------------}
procedure DBF.closedatabase;
{ Frees the memory allocated for database information.                  }

var a:byte;
begin
for a:=1 to nFields
do dispose(Structure[a]);
end;

{-----------------------------------------------------------------------}
function DBF.eof:boolean;
{ Returns true if current record is passed end of file.                 }

begin
eof:= self.recno=-1;
end;
{-----------------------------------------------------------------------}
function DBF.bof:boolean;
{ Returns true if current record is before the begin of file.           }

begin
bof:= self.recno=0;
end;

{-----------------------------------------------------------------------}
function DBF.go(registro:longint):longint;
{ Moves current record pointer to the indicated recird.                 }

begin
if registro>self.enc.cantReg then registro:=-1;{DBEOF}
if registro<1 then registro:=0 {DBBOF};
self.recno:=registro;
go:=registro;
end;

{-----------------------------------------------------------------------}
procedure DBF.gotop;
{ Moves current record pointer to the First reciord.                    }

begin
if self.enc.cantreg>0
then self.recno:=1
else self.recno:=-1;{DBEOF}
end;

{-----------------------------------------------------------------------}
procedure DBF.gobottom;
{ Moves current record pointer to the Last record.                      }

begin
if self.enc.cantreg>0
then self.recno:=self.enc.cantreg
else self.recno:=-1;{DBEOF}
end;

{-----------------------------------------------------------------------}
function DBF.skip(registros:longint):longint;
{ Moves current record <registros> records from current record.         }
{                             (the value may be positive or negative)   } 

begin
inc(self.recno,registros);

if self.recno>self.enc.cantReg
then self.recno:=-1 {DBEOF}
else if self.recno<1
     then self.recno:=0 {BOF};
skip:=self.recno;
end;
{$ifdef fpc}
{-----------------------------------------------------------------------}
function DBF.skip:longint;
begin
skip:=self.skip(1);
end;
{$endif}


{-----------------------------------------------------------------------}
{ devuelve una cadena con el formato y longitud indicados               }
function format(cadena:string; 
                formato:string; 
		longitud:integer;
                decimales:integer):string;
{Formats a string for the specified type, length and precision}

var a: integer;
    s: string;
begin
     s:='';
     case formato[1]
     of 'N': begin
             {habria que verificar el overflow}
             end;
        'C': begin
             s:=cadena;
             for a:=length(cadena)+1 to longitud
             do s:=s+' ';
             s:=copy(s,1,longitud);
             end;
     end;
     format:=s;
end;



{$ifdef fpc}
{-----------------------------------------------------------------------}
function dbf.leer(campo:string; var dato_string:string):integer;
{ Retrieves the field <campo> from the curent record
            into the variable <dato>
The data type of <dato> must be
         String for character date fields.
	 Real for numeric fields.
	 Boolean for Logical fields.
                                                                         } 

var f:file;
    i:integer;
    a:integer;
    Ncampo:byte;

    encontrado:boolean;
    buffer:array[0..4095] of char;
    cadena:string;

begin

     nCampo:=1;
     encontrado:=false;
     while (nCampo<=nFields) and not encontrado
     do begin
        encontrado:=structure[nCampo]^.nam=campo;
        inc(nCampo);
        end;
     dec(Ncampo);
     if not encontrado
     then begin
          leer:=-1;
          exit;
          end;

     ultcampo:=structure[nCampo]^;
          
     {$I-}
     assign(f,self.archivo);
     reset(f,1);
     i:=ioresult;
     if i>0
     then begin
          leer:=i;
          exit;
          end;

     {localizar registro}

     seek(f,self.enc.longEnc + self.enc.longreg*(self.recno-1));

     blockread(f,buffer,self.enc.longReg);

     self.deleted:= buffer[0]='*';

     cadena:='';
     for a:=0 to structure[nCampo]^.len-1
     do cadena:=cadena+buffer[structure[nCampo]^.off+a];
     
     dato_string:=rtrim(cadena);

     leer:=0;
     close(f);
     {$I+}

end;
{-----------------------------------------------------------------------}
function dbf.leer(campo:string):string;
var a:string;
begin
leer(campo,a);
leer:=a;
end;

{-----------------------------------------------------------------------}
function dbf.leer(campo:string;var dato_longint:longint):integer;

var
    cadena:string;
    a,i:integer;

begin

a:=dbf.leer(campo,cadena);
dato_longint:=0;
if a=0
then begin
     if ultcampo.dec>0
     then begin
          cadena:=copy(cadena,1,length(cadena)-ultcampo.dec-1)
                 +copy(cadena,length(cadena)-ultcampo.dec+1,ultcampo.dec);
          end;
     val(cadena,dato_longint,i);
     end;

leer:=a;

end;

{-----------------------------------------------------------------------}
function dbf.leer(campo:string;var dato_real:real):integer;


var
    cadena:string;
    a,i:integer;

begin

a:=dbf.leer(campo,cadena);
dato_real:=0;
if a=0
then val(cadena,dato_real,i);

leer:=a;

end;


{-----------------------------------------------------------------------}
function dbf.leer(campo:string;var dato_bool:boolean):integer;

var
    cadena:string;
    a:integer;

begin

a:=dbf.leer(campo,cadena);
dato_bool:=false;
if a=0
then dato_bool:= ((cadena='Y') or (cadena='T'));

leer:=a;

end;



{-----------------------------------------------------------------------}
function dbf.grabar(campo:string;dato_string:string):integer;
{ Writes the data in the <variable> to the field specified by <campo> 
          of the current record.
The data type of <dato> must be
         String for character and date fields.
	 Real for numeric fields.
	 Boolean for Logical fields.
                                                                         } 
var f:file;
    i:integer;
    Ncampo:byte;

    encontrado:boolean;
    cadena:string;
    dato_real:real;
begin
     grabar:=-1;
     if self.eof then exit;

     nCampo:=1;
     encontrado:=false;
     while (nCampo<=nFields) and not encontrado
     do begin
        encontrado:=structure[nCampo]^.nam=campo;
        inc(nCampo);
        end;
     dec(Ncampo);
     if not encontrado
     then begin
          grabar:=-1;
          exit;
          end;

     {$I-}
     assign(f,self.archivo);
     reset(f,1);
     i:=ioresult;
     if i>0
     then begin
          grabar:=i;
          exit;
          end;

     {De acuerdo al tipo de dato, armo la cadena}

     val(dato_string,dato_real,i);
     
     case structure[nCampo]^.typ
     of 'N': str(dato_real:structure[nCampo]^.len:structure[nCampo]^.dec,cadena);
        'M': str(dato_real:structure[nCampo]^.len,cadena);
        'F': begin
             cadena[0]:=chr(structure[nCampo]^.len);
             move(dato_real,cadena[1],structure[nCampo]^.len);
             end;
        'C': cadena:=format(dato_string,'C',structure[nCampo]^.len,0);
        'D': cadena:=format(dato_string,'C',structure[nCampo]^.len,0);
        'L': if (dato_string='T') or (dato_string='Y')
             or (dato_string='t') or (dato_string='y')
             or (dato_string='1')
             then cadena:='T'
             else cadena:='F'
        end;



{     move(cadena[1],buffer[structure[nCampo]^.off],structure[nCampo]^.len);}

     seek(f,self.enc.longEnc + self.enc.longreg*(self.recno-1)+structure[nCampo]^.off);
     blockwrite(f,cadena[1],structure[nCampo]^.len);

     grabar:=0;
     close(f);
     {$I+}

end;


{-----------------------------------------------------------------------}
function dbf.grabar(campo:string;dato_real:real):integer;
var
    cadena:string;
begin
str(dato_real,cadena);
grabar:=dbf.grabar(campo,cadena);
end;

{-----------------------------------------------------------------------}
function dbf.grabar(campo:string;dato_longint:longint):integer;
var
    cadena:string;
begin
str(dato_longint,cadena);
grabar:=dbf.grabar(campo,cadena);
end;

{-----------------------------------------------------------------------}
function dbf.grabar(campo:string;dato_bool:boolean):integer;
var
    cadena:string;
begin
if dato_bool
then cadena:='T'
else cadena:='F';
grabar:=dbf.grabar(campo,cadena);
end;


{-----------------------------------------------------------------------}
{ Busca el primer registro que contenga el campo indicado comenzando
  por el registro actual.
  Devuelve el numero de registro, o en caso de error: dbEOF (si no se
  encuentra, o el codigo de ioresult multiplicado por (-1).
    
}

function dbf.locate(campo:string;dato_string:string;valorskip:longint):integer;
{ Searchs in the database for a record containing a fiald <campo> with the
          data specified in <variable> .
  Returns the recordnumber or:
	 dbEOF if it was not found,
	 [(-1)*ioresult] if it was a file error.
  The current record stays pointing to the found record.
  The data type of <dato> must be
         String for character and date fields.
	 Real for numeric fields.
	 Boolean for Logical fields.
                                                                         } 

var f:file;
    i:integer;
    a:integer;
    Ncampo:byte;

    encontrado:boolean;
    buffer:array[0..4095] of char;
    cadena:string;
    cadena1:string;

    dato_real:real;



begin


     found:=false;


     nCampo:=1;
     encontrado:=false;
     while (nCampo<=nFields) and not encontrado
     do begin
        encontrado:=structure[nCampo]^.nam=campo;
        inc(nCampo);
        end;
     dec(Ncampo);
     if not encontrado
     then begin
          locate:=DBEOF;
	  found:=false;
          exit;
          end;

     val(dato_string,dato_real,i);

     case structure[nCampo]^.typ
     of 'N': str(dato_real:structure[nCampo]^.len:structure[nCampo]^.dec,cadena);
        'M': str(dato_real:structure[nCampo]^.len,cadena);
        'F': begin
             cadena[0]:=chr(structure[nCampo]^.len);
             move(dato_real,cadena[1],structure[nCampo]^.len);
             end;
        'C': cadena:=format(dato_string,'C',structure[nCampo]^.len,0);
        'D': cadena:=format(dato_string,'C',structure[nCampo]^.len,0);
        'L': if (dato_string='T') or (dato_string='Y')
             or (dato_string='t') or (dato_string='y')
             or (dato_string='1')
             then cadena:='T'
             else cadena:='F'
        end;


     {$I-}
     assign(f,self.archivo);
     reset(f,1);
     i:=ioresult;
     if i>0
     then begin
          locate:=-i;
	  found:=false;
          exit;
          end;

     {localizar registro}

     seek(f,self.enc.longEnc + self.enc.longreg*(self.recno-1));

     encontrado:=false;
     repeat
         blockread(f,buffer,self.enc.longReg);

         cadena1:='';
         for a:=0 to structure[nCampo]^.len-1
         do cadena1:=cadena1+buffer[structure[nCampo]^.off+a];

         encontrado:=cadena=cadena1;

         if not encontrado
         then skip(valorskip);

     until encontrado or (recno= -1{DBEOF}) or (recno= 0{DBBOF});

     self.deleted:= buffer[0]='*';

     locate:= recno;
     found:= recno<>-1;
     close(f);
     {$I+}

end;

{-----------------------------------------------------------------------}
function dbf.locate(campo:string;dato_string:string):integer;

begin
locate:=self.locate(campo,dato_string,1);
end;

{-----------------------------------------------------------------------}
function dbf.locate(campo:string;dato_real:real):integer;
var
    cadena:string;
begin
str(dato_real,cadena);
locate:=dbf.locate(campo,cadena,1);
end;

{-----------------------------------------------------------------------}
function dbf.locate(campo:string;dato_real:real;valorskip:longint):integer;
var
    cadena:string;
begin
str(dato_real,cadena);
locate:=dbf.locate(campo,cadena,valorskip);
end;

{-----------------------------------------------------------------------}
function dbf.locate(campo:string;dato_longint:longint):integer;
var
    cadena:string;
begin
str(dato_longint,cadena);
locate:=dbf.locate(campo,cadena,1);
end;

{-----------------------------------------------------------------------}
function dbf.locate(campo:string;dato_longint:longint;valorskip:longint):integer;
var
    cadena:string;
begin
str(dato_longint,cadena);
locate:=dbf.locate(campo,cadena,valorskip);
end;

{-----------------------------------------------------------------------}
function dbf.locate(campo:string;dato_bool:boolean):integer;
var
    cadena:string;
begin
if dato_bool
then cadena:='T'
else cadena:='F';
locate:=dbf.locate(campo,cadena,1);
end;

{-----------------------------------------------------------------------}
function dbf.locate(campo:string;dato_bool:boolean; valorskip:longint):integer;
var
    cadena:string;
begin
if dato_bool
then cadena:='T'
else cadena:='F';
locate:=dbf.locate(campo,cadena,valorskip);
end;


{$else}


{-----------------------------------------------------------------------}
function dbf.leer(campo:string;var dato):integer;
{ Retrieves the field <campo> from the curent record
            into the variable <dato>
The data type of <dato> must be
         String for character date fields.
	 Real for numeric fields.
	 Boolean for Logical fields.
                                                                         } 

var f:file;
    i:integer;
    a:integer;
    Ncampo:byte;

    encontrado:boolean;
    buffer:array[0..4095] of char;
    cadena:string;

    dato_long:^longint;
    dato_string:^string;
    dato_real:^real;
    dato_boolean:^boolean;



begin

     dato_long:=@dato;
     dato_string:=@dato;
     dato_real:=@dato;
     dato_boolean:=@dato;

     nCampo:=1;
     encontrado:=false;
     while (nCampo<=nFields) and not encontrado
     do begin
        encontrado:=structure[nCampo]^.nam=campo;
        inc(nCampo);
        end;
     dec(Ncampo);
     if not encontrado
     then begin
          leer:=-1;
          exit;
          end;

     {$I-}
     assign(f,self.archivo);
     reset(f,1);
     i:=ioresult;
     if i>0
     then begin
          leer:=i;
          exit;
          end;

     {localizar registro}

     seek(f,self.enc.longEnc + self.enc.longreg*(self.recno-1));

     blockread(f,buffer,self.enc.longReg);

     self.deleted:= buffer[0]='*';

     cadena:='';
     for a:=0 to structure[nCampo]^.len-1
     do cadena:=cadena+buffer[structure[nCampo]^.off+a];



     case structure[nCampo]^.typ
     of 'N','M','F': val(cadena,dato_real^,i);
        'C','D': dato_string^:=rtrim(cadena);
        'L': dato_boolean^:= ((cadena='Y') or (cadena='T'));
        end;

     leer:=0;
     close(f);
     {$I+}

end;
{-----------------------------------------------------------------------}
function dbf.grabar(campo:string;var dato):integer;
{ Writes the data in the <variable> to the field specified by <campo> 
          of the current record.
The data type of <dato> must be
         String for character and date fields.
	 Real for numeric fields.
	 Boolean for Logical fields.
                                                                         } 

var f:file;
    i:integer;
    Ncampo:byte;

    encontrado:boolean;
    cadena:string;

    dato_long:^longint;
    dato_string:^string;
    dato_real:^real;
    dato_boolean:^boolean;


begin
     grabar:=-1;
     if self.eof then exit;

     dato_long:=@dato;
     dato_string:=@dato;
     dato_real:=@dato;
     dato_boolean:=@dato;

     nCampo:=1;
     encontrado:=false;
     while (nCampo<=nFields) and not encontrado
     do begin
        encontrado:=structure[nCampo]^.nam=campo;
        inc(nCampo);
        end;
     dec(Ncampo);
     if not encontrado
     then begin
          grabar:=-1;
          exit;
          end;

     {$I-}
     assign(f,self.archivo);
     reset(f,1);
     i:=ioresult;
     if i>0
     then begin
          grabar:=i;
          exit;
          end;



     {De acuerdo al tipo de dato, armo la cadena}

     case structure[nCampo]^.typ
     of 'N': str(dato_real^:structure[nCampo]^.len:structure[nCampo]^.dec,cadena);
        'M': str(dato_real^:structure[nCampo]^.len,cadena);
        'F': begin
             cadena[0]:=chr(structure[nCampo]^.len);
             move(dato_real,cadena[1],structure[nCampo]^.len);
             end;
        'C': cadena:=format(dato_string^,'C',structure[nCampo]^.len,0);
        'D': cadena:=format(dato_string^,'C',structure[nCampo]^.len,0);
        'L': if dato_boolean^ = true
             then cadena:='T'
             else cadena:='F'
        end;


{     move(cadena[1],buffer[structure[nCampo]^.off],structure[nCampo]^.len);}

     seek(f,self.enc.longEnc + self.enc.longreg*(self.recno-1)+structure[nCampo]^.off);
     blockwrite(f,cadena[1],structure[nCampo]^.len);

     grabar:=0;
     close(f);
     {$I+}

end;

{-----------------------------------------------------------------------}
{ Busca el primer registro que contenga el campo indicado comenzando
  por el registro actual.
  Devuelve el numero de registro, o en caso de error: dbEOF (si no se 
  encuentra, o el codigo de ioresult multiplicado por (-1).
}

function dbf.locate(campo:string;var dato;valorskip:longint):integer;
{ Searchs in the database for a record containing a fiald <campo> with the
          data specified in <variable> .
  Returns the recordnumber or:
	 dbEOF if it was not found,
	 [(-1)*ioresult] if it was a file error.
  The current record stays pointing to the found record.
  The data type of <dato> must be
         String for character and date fields.
	 Real for numeric fields.
	 Boolean for Logical fields.
                                                                         } 

var f:file;
    i:integer;
    a:integer;
    Ncampo:byte;

    encontrado:boolean;
    buffer:array[0..4095] of char;
    cadena:string;
    cadena1:string;

    dato_long:^longint;
    dato_string:^string;
    dato_real:^real;
    dato_boolean:^boolean;



begin

     dato_long:=@dato;
     dato_string:=@dato;
     dato_real:=@dato;
     dato_boolean:=@dato;

     found:=false;


     nCampo:=1;
     encontrado:=false;
     while (nCampo<=nFields) and not encontrado
     do begin
        encontrado:=structure[nCampo]^.nam=campo;
        inc(nCampo);
        end;
     dec(Ncampo);
     if not encontrado
     then begin
          locate:=DBEOF;
	  found:=false;
          exit;
          end;


     case structure[nCampo]^.typ
     of 'N': str(dato_real^:structure[nCampo]^.len:structure[nCampo]^.dec,cadena);
        'M': str(dato_real^:structure[nCampo]^.len,cadena);
        'F': begin
             cadena[0]:=chr(structure[nCampo]^.len);
             move(dato_real,cadena[1],structure[nCampo]^.len);
             end;
        'C': cadena:=format(dato_string^,'C',structure[nCampo]^.len,0);
        'D': cadena:=format(dato_string^,'C',structure[nCampo]^.len,0);
        'L': if dato_boolean^ = true
             then cadena:='T'
             else cadena:='F'
        end;


     {$I-}
     assign(f,self.archivo);
     reset(f,1);
     i:=ioresult;
     if i>0
     then begin
          locate:=-i;
	  found:=false;
          exit;
          end;

     {localizar registro}

     seek(f,self.enc.longEnc + self.enc.longreg*(self.recno-1));

     encontrado:=false;
     repeat
         blockread(f,buffer,self.enc.longReg);

         cadena1:='';
         for a:=0 to structure[nCampo]^.len-1
         do cadena1:=cadena1+buffer[structure[nCampo]^.off+a];

         encontrado:=cadena=cadena1;

         if not encontrado
         then skip(valorskip);

     until encontrado or (recno= -1{DBEOF})or (recno= 0{DBBOF});

     self.deleted:= buffer[0]='*';

     locate:=recno;
     found:= recno<>-1;
     close(f);
     {$I+}

end;

{$endif}




{-----------------------------------------------------------------------}
function dbf.delete_mark(marca:char):integer;
{Writes a delete mark specified by <marca> in the current record.
<marca> should be '*' for deleted and ' ' for not deleted. Other values
are not validated but the result may be unpredictible when the database
is readed by other programs (i.e. foxbase)                              }

var f:file;
    i:integer;

begin

     {$I-}
     assign(f,self.archivo);
     reset(f,1);
     i:=ioresult;
     if i>0
     then begin
          delete_mark:=i;
          exit;
          end;

     seek(f,self.enc.longEnc + self.enc.longreg*(self.recno-1));
     blockwrite(f,marca,1);

     delete_mark:=0;
     close(f);
     {$I+}

end;

{-----------------------------------------------------------------------}
function dbf.delete:integer;
{Writes a delete mark into the current record.                          }

begin
delete:=delete_mark('*');
end;
{-----------------------------------------------------------------------}
function dbf.recall:integer;
{Removes the delete mark of the current record.                         }

begin
recall:=delete_mark(' ');
end;
{-----------------------------------------------------------------------}
function dbf.appendblank:integer;
{Appends a blank record to the database                                 }

var f:file;
    i:integer;
    a:word;
    buffer:array[0..4095] of char;

begin

     {$I-}
     assign(f,self.archivo);
     reset(f,1);
     i:=ioresult;
     if i>0
     then begin
          appendblank:=i;
          exit;
          end;

     self.recno:=self.enc.cantreg+1;

     seek(f,self.enc.longEnc + self.enc.longreg*(self.recno-1));
     for a:=1 to self.enc.longReg
     do buffer[a]:=' ';

     blockwrite(f,buffer,self.enc.longReg);
     self.enc.cantreg:=self.recno;
     seek(f,0);
     blockwrite(f,self.enc,sizeof(self.enc));

     appendblank:=0;
     close(f);
     {$I+}

end;

{-----------------------------------------------------------------------}
function dbf.reccount:longint;
{Returns the number of records from current table, including deleted.   }

begin
    reccount:=self.enc.cantreg;
end;

{-----------------------------------------------------------------------}

begin
end.
